# example: A ring-loaded Mach-Zehnder interferometer.
# A directiopnal coupler is used for splitter and combiner
# and each MZI arm has a ring in it.
from ipkiss3 import all as i3
from ring import RingResonator
from dircoup import BentDirectionalCoupler


TECH = i3.TECH


class RingLoadedMZI(i3.PCell):

    """A Mach-Zehnder with a ring resonator in each arm"""

    _name_prefix = "MZI"
    # 1. Define the rings and splitters as childcells.
    # We add a type restriction so ensure the passed class inherits from RingResonator/BentDirectionalCoupler.

    ring1 = i3.ChildCellProperty(restriction=i3.RestrictType(RingResonator))
    ring2 = i3.ChildCellProperty(restriction=i3.RestrictType(RingResonator))
    splitter = i3.ChildCellProperty(restriction=i3.RestrictType(BentDirectionalCoupler))
    combiner = i3.ChildCellProperty(restriction=i3.RestrictType(BentDirectionalCoupler))

    # 2. We define a waveguide template and the waveguide cells
    wg_template = i3.WaveguideTemplateProperty(default=i3.TECH.PCELLS.WG.DEFAULT)
    wgs = i3.ChildCellListProperty(doc="list of waveguides")

    # 3. We first define the 4 rounded waveguides that we need at the PCELL level.
    def _default_wgs(self):
        wgs = []
        for i in range(4):
            wg = i3.RoundedWaveguide(name=self.name + "_w" + str(i + 1), trace_template=self.wg_template)
            wgs.append(wg)
        return wgs

    class Layout(i3.LayoutView):
        bend_radius = i3.PositiveNumberProperty(default=i3.TECH.WG.BEND_RADIUS)

        # 4. Placing the components
        @i3.cache()
        def _get_components(self):

            si_splitter = self.splitter.size_info()  # object with contours of the component
            si_combiner = self.combiner.size_info()  # object with contours of the component
            si_ring1 = self.ring1.size_info()  # object with contours of the component
            si_ring2 = self.ring2.size_info()  # object with contours of the component

            # spacing between components to allow easy routing (can be optimized)
            spacing = 3 * self.bend_radius + 3 * TECH.WG.SHORT_STRAIGHT

            # splitter
            t_splitter = i3.IdentityTransform()  # place the splitter in (0,0)
            splitter = i3.SRef(reference=self.splitter, name="splitter", transformation=t_splitter)

            # ring1
            t_ring1 = i3.Translation(
                (si_splitter.east - si_ring1.west + spacing, si_splitter.north - si_ring1.south + spacing)
            )
            ring1 = i3.SRef(reference=self.ring1, name="ring1", transformation=t_ring1)

            # ring2
            t_ring2 = i3.VMirror() + i3.Translation(
                (si_splitter.east - si_ring2.west + spacing, si_splitter.south + si_ring2.south - spacing)
            )
            ring2 = i3.SRef(reference=self.ring2, name="ring2", transformation=t_ring2)

            # combiner
            si_rings = si_ring1.transform(t_ring1) + si_ring2.transform(t_ring2)
            t_combiner = i3.Translation((si_rings.east - si_combiner.west + spacing, 0.0))
            combiner = i3.SRef(reference=self.combiner, name="combiner", transformation=t_combiner)

            return splitter, ring1, ring2, combiner

        # 5. Getting the layout view of the waveguides.
        def _default_wgs(self):

            splitter, ring1, ring2, combiner = self._get_components()
            wg_cells = self.cell.wgs
            wg_template = self.wg_template

            # Defining the connections between ports using a tuple shorthand
            connections = [
                (splitter.ports["out_1"], ring1.ports["in"]),  # splitter to ring1
                (combiner.ports["in_1"], ring1.ports["out"]),  # combiner to ring1
                (splitter.ports["out_2"], ring2.ports["in"]),  # splitter to ring2
                (combiner.ports["in_2"], ring2.ports["out"]),  # combiner to ring2
            ]

            # Updating the 4 layout views of the waveguides using a loop
            wg_layouts = []
            for wg_cell, connection in zip(wg_cells, connections):
                wg_layout = wg_cell.get_default_view(i3.LayoutView)
                wg_layout.set(
                    bend_radius=self.bend_radius,
                    shape=i3.RouteManhattan(
                        start_port=connection[0], end_port=connection[1], bend_radius=self.bend_radius
                    ),
                    trace_template=wg_template,
                )

                wg_layouts.append(wg_layout)

            return wg_layouts

        # 6. Placing both the component and the waveguides in the layout.
        def _generate_instances(self, insts):

            insts += self._get_components()

            for i, wg_layout in enumerate(self.wgs):
                insts += i3.SRef(reference=wg_layout, name="w" + str(i))

            return insts

        # 7. Generating the ports of the rlmzi
        def _generate_ports(self, prts):
            instances = self.instances
            prts += instances["splitter"].ports["in_1"]
            prts += instances["splitter"].ports["in_2"]
            prts += instances["combiner"].ports["out_1"]
            prts += instances["combiner"].ports["out_2"]
            return prts
